static const char target[]               = "x86_64-pc-linux-gnu";
static const char *const startfiles[]    = {"-l", ":crt1.o", "-l", ":crti.o", "-l", ":crtbegin.o"};
static const char *const endfiles[]      = {"-l", "c", "-l", ":crtend.o", "-l", ":crtn.o"};
static const char *const preprocesscmd[] = {
	"/usr/bin/cpp", "-undef",

	/* we don't yet support these optional features */
	"-D", "__STDC_NO_ATOMICS__",
	"-D", "__STDC_NO_COMPLEX__",

	/* ignore extension markers */
	"-D", "__extension__=",

	"-D", "__STDC_ENDIAN_LITTLE__=1234",
	"-D", "__STDC_ENDIAN_BIG__=4321",
	"-D", "__STDC_ENDIAN_NATIVE__=__STDC_ENDIAN_LITTLE__",
	"-D", "__ORDER_LITTLE_ENDIAN__=__STDC_ENDIAN_LITTLE__",
	"-D", "__ORDER_BIG_ENDIAN__=__STDC_ENDIAN_BIG__",
	"-D", "__BYTE_ORDER__=__STDC_ENDIAN_NATIVE__",
	"-D", "__FLT_EVAL_METHOD__=1",
	"-D", "__BITINT_MAXWIDTH__=64",

	"-D", "__x86_64__",
	"-D", "__amd64__",

	"-D", "__LP64__",
	"-D", "__SIZEOF_LONG__=8",
	"-D", "__SIZEOF_POINTER__=8",
	"-D", "__SIZEOF_SIZE_T__=8",
	"-D", "__SIZEOF_PTRDIFF_T__=8",

	"-D", "__linux__",
	"-D", "__unix__",
	"-D", "__INT_MAX__=0x7fffffff",
	"-D", "__LONG_LONG_MAX__=0x7fffffffffffffffL",
};
static const char *const codegencmd[]    = {"/usr/bin/qbe", };
static const char *const assemblecmd[]   = {"/usr/bin/as", };
static const char *const linkcmd[]       = {"/usr/bin/ld", "-L", "/usr/lib/gcc/x86_64-pc-linux-gnu/16", "--dynamic-linker", "/lib64/ld-linux-x86-64.so.2"};
